/*

Name:
DRV_DX8.C

Description:
Mikmod driver for output on Windows/DirectX 8

Portability:
DirectX 8

*/
#include <dsound.h>
#ifdef _XBOX
#include <xtl.h>
#else
#include <windows.h>
#endif
#include <process.h>
#include <stdio.h>
#include "mikmod.h"

static IDirectSound8 *ds;
static IDirectSoundBuffer *buffer;
static DWORD block_size, buf_size, cur_offset = 0;


BOOL DX8_IsThere(void)
{
	#ifdef _XBOX
	#define DirectSoundCreate8 DirectSoundCreate
	#endif
	return DirectSoundCreate8( NULL, &ds, NULL ) == DS_OK;
}


BOOL DX8_Init(void)
{
  DSBUFFERDESC dsbdesc; 
   
  md_mode = DMODE_16BITS | DMODE_STEREO;
//  md_mixfreq = 44100;

  // Set up wave format structure. 
  WAVEFORMATEX wf;

  memset(&wf, 0, sizeof(wf)); 
  wf.wFormatTag = WAVE_FORMAT_PCM; 
  wf.nChannels = md_mode & DMODE_STEREO ? 2 : 1; 
  wf.wBitsPerSample = md_mode & DMODE_16BITS ? 16 : 8;
  wf.nSamplesPerSec = md_mixfreq; 
  wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8; 
  wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign; 
 
  // Set up DSBUFFERDESC structure. 
  memset(&dsbdesc, 0, sizeof(DSBUFFERDESC)); 
  dsbdesc.dwSize = sizeof(DSBUFFERDESC); 
  dsbdesc.dwFlags = DSBCAPS_CTRLVOLUME; 
  dsbdesc.dwBufferBytes = 1*wf.nAvgBytesPerSec; 
  dsbdesc.lpwfxFormat = &wf; 

  buf_size = dsbdesc.dwBufferBytes;
  block_size = buf_size / 16; 

  // Create buffer. 
  if( !SUCCEEDED( ds->CreateSoundBuffer( &dsbdesc, &buffer, NULL ) ) ) {
	  return FALSE;
  }

  if( !VC_Init() ) {
	  return 0;
  }

#ifndef _XBOX
  ds->SetCooperativeLevel( GetForegroundWindow(), DSSCL_PRIORITY );
#endif

  return 1;
}


void DX8_Exit(void)
{
	buffer->Stop(); 
	buffer = NULL;
	VC_Exit();
}

void debug( const char* msg )
{
#ifdef _DEBUG
	FILE* f = fopen( "qrash.log", "a" );
	fputs( msg, f );
	fclose( f );
#endif
}

void mix()
{
	if( !buffer ) return;
	DWORD status, play_cursor = 0, write_cursor = 0;
	buffer->GetStatus( &status );
	if( status & DSBSTATUS_PLAYING ) {
		buffer->GetCurrentPosition( &play_cursor, &write_cursor );
	}
	else {
		write_cursor = 0;
	}
#ifdef _DEBUG
	char msg[256];
	sprintf( msg, "play_cursor=%d write_cursor=%d cur_offset=%d\n", 
			 play_cursor, write_cursor, cur_offset );
	debug( msg );
#endif
	if( DWORD( abs( cur_offset - write_cursor ) ) < block_size ) {
		void *bytes1, *bytes2 = 0;
		DWORD size1, size2 = 0;
		if( SUCCEEDED( buffer->Lock( cur_offset, block_size, &bytes1, &size1, &bytes2, &size2, 0 ) ) ) {
#ifdef _DEBUG
			char msg[256];
			sprintf( msg, "cur_offset=%d bytes1=%p size1=%d bytes2=%p size2=%d\n", 
					 cur_offset, bytes1, size1, bytes2, size2 );
			debug( msg );
#endif
			VC_WriteBytes( (SBYTE*)bytes1, (UWORD)size1 );
			if( bytes2 && size2 ) {
				VC_WriteBytes( (SBYTE*)bytes2, (UWORD)size2 );
			}
			cur_offset += block_size;
			cur_offset %= buf_size;
			buffer->Unlock( bytes1, size1, bytes2, size2 );
		}
	}
}

void update_thread( void* unused )
{
	if( !buffer ) {
		_endthread();
	}
	while( 1 ) {
		mix();
		Sleep(25);
	}
}

void DX8_PlayStart(void)
{
	VC_PlayStart();
	buffer->SetVolume( DSBVOLUME_MAX );
	mix();
	HRESULT hr = buffer->Play( 0, 0, DSBPLAY_LOOPING );
	if( SUCCEEDED( hr ) ) {
		_beginthread( &update_thread, 16384, NULL );
	}
}


void DX8_PlayStop(void)
{
	buffer->Stop(); 
	VC_PlayStop();
}


void DX8_Update(void)
{
	/* does nothing, buffers are updated in the background */
}


DRIVER drv_dx8={
	NULL,
	"DirectX 8",
	"MikMod DirectSound8 Driver v1.0",
	DX8_IsThere,
	VC_SampleLoad,
	VC_SampleUnload,
	DX8_Init,
	DX8_Exit,
	DX8_PlayStart,
	DX8_PlayStop,
	DX8_Update,
	VC_VoiceSetVolume,
	VC_VoiceSetFrequency,
	VC_VoiceSetPanning,
	VC_VoicePlay
};

